/*************************************************************************************/
/* Workshop de
 * Supercomputacin y Clculo Cientfico, SC3-2014
 * SC3UIS con Academia NVIDIA CUDA
 
	Este codigo presenta una implementacion en CUDA C de la inversion de las
	posiciones de un vector de valores enteros entre (0 y 255), usando 
	transferencias de memoria global y ejecucion, multithread y multiblock,
	para introducir los conceptos:
	
		* Manejo de memoria en host y device,
		* Copiado y transferencia de variables entre host y device 
		* Invocacion de kernel.
	
 * Creado por: William J Trigos, Ing. Electrnico, UIS(2013) 
 * Fecha: 20/04/2014 12:49 PM, (Bucaramanga, Colombia)
 */
/*************************************************************************************/

#include <stdio.h>
#include <stdlib.h>

/* Declaracion de funciones Utilitarias */
/* Funcion de Utilidad para el chequeo y reporte de errores en CUDA*/
void checkCUDAError(const char*);

/*************************************************************************************/
/* Kernel para negacion de las posiciones de un vector en GPU (d_a) */
/*************************************************************************************/

/* Version Multi-Thread, ejecudada sobre un solo bloque */
__global__ void negate(int *d_a)
{
    /* Parte 2B: negar los elementos de d_a */
}

/* Version Multi-block del kernel para la Parte 2C */
__global__ void negate_multiblock(int *d_a)
{
    /* Parte 2C: negar los elementos de d_a, usando multiples bloques a la vez */
}

/* Numero de elementos tipo enteros en el vector */
#define ARRAY_SIZE 256

/*	RECOMENDACION: 

	El nmero de Bloques CUDA y threads por bloque a utilizar. 
	siempre deben ser multiplos del tamao del vector. 
	de lo contrario, debe hacerse una operacion pertinente para 
	garantizar la ejecucion del kernel sobre los elementos restante
	y evitar la misma sobre elementos del vector que no correspondan
	al problema en cuestion. 
	(Condiciones de carrera y/o conflictos de memoria)
	
   Para el kernel de un solo bloque, NUM_BLOCKS debe ser 1 y 
   THREADS_PER_BLOCK debe ser el tamao del vector
 */
#define NUM_BLOCKS  1
#define THREADS_PER_BLOCK 256

/* Programa Principal */
int main(int argc, char *argv[])
{
    int *h_a, *h_out;
    int *d_a;

    int i;
    size_t sz = ARRAY_SIZE * sizeof(int);

    /*
	Reserva de memoria para los vectores en el host
		* h_a vector de entrada
		* h_out vector para los resultados
    */
    h_a = (int *) malloc(sz);
    h_out = (int *) malloc(sz);

    /* Reserva de memoria sobre la memoria global de la GPU para el vector d_a */
    /* Parte 1A: reserva de memoria en GPU para d_a */
	cudaMalloc( );

    /* Llenado del vector de entrada h_a e inicializacion a cero de h_out en el host */
    for (i = 0; i < ARRAY_SIZE; i++) {
        h_a[i] = i;
        h_out[i] = 0;
		/* Impresion del vector de entrada en la terminal */
		printf("%d, ", h_a[i]);
    }

    printf("\n\n");

    /* Copia h_a en el host hacia d_a en la memoria global GPU */
	/* Part 1B: copia del vector h_a en el host hacia el vector d_a en GPU */
    cudaMemcpy( );

    /* declaraciones de la grid y los bloques, para el lanzamiento del kernel negate */
    /* Parte 2A: configurar la ejecucion de un kernel (decomente y complete las lineas siguientes) */
    // dim3 blocksPerGrid( ); 
    // dim3 threadsPerBlock( );
    // negate<<< , >>>( );

    /* Sincronizacion: esperar a que todos los threads completen las operaciones */
    cudaThreadSynchronize();
	/*  Chequeo y reporte de errores CUDA, despues de la ejecucion del kernel */
    checkCUDAError("kernel invocation");

    /* Copia de los resultados obtenidos en la GPU hacia el host, d_a --> h_out */
    /* Parte 1C: copiar el vector d_a hacia el vector h_out en el host */
	cudaMemcpy( );
	/*  Chequeo y reporte de errores CUDA, despues de la ejecucion de cudaMemcpy */
    checkCUDAError("memcpy");

    /* Impresion de los resultados en la terminal */
    printf("Results: ");
    for (i = 0; i < ARRAY_SIZE; i++) {
      printf("%d, ", h_out[i]);
    }
    printf("\n\n");

    /* liberacion de memoria del vector en GPU (d_a)*/
    /* Parte 1D: libere memoria de la GPU usada para d_a */
    cudaFree( );

    /* liberacion de memoria de los vectores en CPU (h_a, h_out)*/
    free(h_a);
    free(h_out);

    return 0;
}

/* Funcion de Utilidad para el chequeo y reporte de errores en CUDA*/
void checkCUDAError(const char *msg)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {
        fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
    }                         
}
