/*************************************************************************************/
/* Workshop de
 * Supercomputacin y Clculo Cientfico, SC3-2014
 * SC3UIS con Academia NVIDIA CUDA
 
	Este codigo presenta la invocacion de un conjunto de kernels CUDA C del metodo 
	Montecarlo para el calculo del numero pi. Los kernels aqui propuestos emplean 
	ejecucion multi-block.
	
	Se implementa una version secuencial del calculo del numero pi, sobre la CPU 
	para el posterior medicion del desempeo relativo del mismo con respecto a las
	versiones paralelas implementadas en los kernels de las librerias Thrust y CUDA C.
	
 * Creado por: William J Trigos, Ing. Electrnico, UIS(2013) 
 * Fecha: 20/04/2014 12:49 PM, (Bucaramanga, Colombia)

Sentencia para la compilacion y ejecucion

	nvcc kernel.cu -o pi -lcurand -I /usr/local/cuda/samples/common/inc
	./pi
/*************************************************************************************/

#define PI 3.14159265359  // Valor de pi segun google =P

//#include "pi.h"
/***************************************************************************************/
/* Cabeceros para C++ */
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

/* Cabeceros para CUDA */
#include <cuda.h>
#include <cuda_runtime.h>

/* Cabeceros para Thrust */
#include <thrust/random.h>
#include <thrust/count.h>
#include <thrust/reduce.h>
#include <thrust/remove.h>
#include <thrust/device_free.h>
#include <thrust/device_malloc.h>
#include <thrust/device_ptr.h>
#include <thrust/transform_reduce.h>

// Para el calculo de pi por el metodo de montecarlo se requieren cerca de 5'000.000 de iteraciones
#define samples 900000000
//static int samples = 2000;//128 * 1024 * 1024;	//   134'217.728 cantidad de puntos, mas precision Montecarlo

float pi_cuda(unsigned int tileSize, unsigned int threadsPerIteration);
float pi_cuda2();
/***************************************************************************************/

#include "pi.cu"		// fichero que contiene los kernels CUDA para el calculo de Pi, por el metodo Monte Carlo

static float pi_cpu(){
    int count = 0;
    for (int i = 0; i < samples; ++i) {
        float x = float(rand()) / RAND_MAX;
        float y = float(rand()) / RAND_MAX;
        if (sqrt(x*x + y*y) < 1)
            count++;
    }
    return 4.0 * count / samples;
}

static void pi_cuda_init(){
	cudaDeviceReset();
}

// Variables Globales necesarias
unsigned int ThreadsPerBlock = 128;	//512;
unsigned int MaxThreadsPerIteration = 128;	//512;

int getMilisegundos(clock_t c){
	int tiempo=0;
	tiempo = (int)((c/(float)CLOCKS_PER_SEC)*1000) ;
	return tiempo;
}

int main(int argc, char* argv[]){

int t0,t1,t2,t3,t4,t5,t6,t7,t8,t9;

		int devCount;
		cudaGetDeviceCount(&devCount);
		// Obteniendo caracteristicas de la GPU
		cudaDeviceProp devProp;
		cudaGetDeviceProperties(&devProp, 0);
		ThreadsPerBlock = devProp.maxThreadsPerBlock;
		MaxThreadsPerIteration = ThreadsPerBlock * ThreadsPerBlock;//devProp.maxThreadsDim[0] * devProp.maxThreadsDim[1];
		printf("CUDA Threads x Block de la GPU \"cuda\" = %d\n", ThreadsPerBlock);
		//printf("CUDA Max Threads x Iteracion Usados por \"cuda\" = %u\n", MaxThreadsPerIteration);
		printf("# de Puntos (samples) :  %d\n", samples);
	        
/* version CPU */		
		t0 = clock();
		float pi_1 = pi_cpu();
		t1 = clock();
		printf("\nPi Version CPU:\n\tpi=  %.12f\terror=  %.3e", pi_1, pi_1 - PI);
		printf("\tTiempo:  %d [ms]\n", getMilisegundos(t1-t0));
 
/* version ArrayFire */	        
	//printf("af      :  %f seconds to estimate  pi=%.12f\n", timeit(wrap_af),   pi_af());

/* version GPU CUDA Thrust 1 */	
	    pi_cuda_init();	// Inicializar las targetas para trabajo de computo CUDA
	    t2 = clock();
		float pi_2 = pi_cuda(ThreadsPerBlock, MaxThreadsPerIteration);
		t3 = clock();
		printf("Pi Version GPU_Thrust_Basico:\n\tpi=  %.12f\terror=  %.3e", pi_2, pi_2 - PI);
		printf("\tTiempo:  %d [ms]\n", getMilisegundos(t3-t2));
		
/* version GPU Thrust Library 2 */
		pi_cuda_init();	// Inicializar las targetas para trabajo de computo CUDA
	    t4 = clock();
		float pi_3 = pi_cuda2();
		t5 = clock();
		printf("Pi Version GPU_Thrust_Transform:\n\tpi=  %.12f\terror=  %.3e", pi_3, pi_3 - PI);
		printf("\tTiempo:  %d [ms]\n", getMilisegundos(t5-t4));
		
		pi_cuda_init();	// Inicializar las targetas para trabajo de computo CUDA

/* version GPU CUDA_Neto*/
		pi_cuda_init();	// Inicializar las targetas para trabajo de computo CUDA
	    t6 = clock();
		float pi_4 = pi_cuda3(ThreadsPerBlock, MaxThreadsPerIteration);
		t7 = clock();
		printf("Pi Version GPU_CUDA_NETO:\n\tpi=  %.12f\terror=  %.3e", pi_4, pi_4 - PI);
		printf("\tTiempo:  %d [ms]\n", getMilisegundos(t7-t6));
		
		pi_cuda_init();	// Inicializar las targetas para trabajo de computo CUDA

/* version GPU CUDA_Neto2*/
/*
		pi_cuda_init();	// Inicializar las targetas para trabajo de computo CUDA
	    t8 = clock();
		float pi_5 = pi_cuda5(ThreadsPerBlock, MaxThreadsPerIteration);
		t9 = clock();
		printf("Pi Version GPU_CUDA_mejor:\n\tpi=  %.12f\terror=  %.3e", pi_5, pi_5 - PI);
		printf("\tTiempo:  %d [ms]\n", getMilisegundos(t9-t8));
		
		pi_cuda_init();	// Inicializar las targetas para trabajo de computo CUDA

*/
	    #ifdef WIN32 // comando de pause en Windows
	    if (!(argc == 2 && argv[1][0] == '-')) {
	        printf("Para Salir [enter]...");
	        getchar();
	    }
	    #endif
	    return 0;
}
